@@ -28,7 +28,8 @@ DEALINGS IN THE SOFTWARE.  */
 #include <string.h>
 #include <stdint.h>
 #include <inttypes.h>
-#include "htslib/hts_endian.h"
+
+#include "../htslib/hts_endian.h"
 
 typedef struct {
     uint8_t u8[2];
